import bpy
from io_scene_mdl2022.mdl_utils_2022 import main_export
from io_scene_mdl2022.mdl_utils_2022 import main_import
from io_scene_mdl2022.mdl_utils_2022 import main_import_20
from bpy_extras.io_utils import ExportHelper
from bpy.props import StringProperty, BoolProperty, EnumProperty
from bpy_extras.io_utils import ImportHelper
from bpy.types import Operator


def read_some_data_20(context, filepath, keywords):
    print( filepath )
    try:
        bpy.ops.object.mode_set(mode='OBJECT')
    except:
        pass
    main_import_20.importMDL( filepath,keywords)
    try:
        bpy.ops.object.mode_set(mode='OBJECT')
    except:
        pass
    return {'FINISHED'}

def write_some_data(context, filepath, keywords):
    print( filepath )
    bpy.ops.object.mode_set(mode='OBJECT')
    main_export.mainExportMDL( filepath , keywords )
    bpy.ops.object.mode_set(mode='OBJECT')
    return {'FINISHED'}

def read_some_data(context, filepath , keywords):
    print( filepath )
    try:
        bpy.ops.object.mode_set(mode='OBJECT')
    except:
        pass
    main_import.importMDL( filepath , keywords)
    try:
        bpy.ops.object.mode_set(mode='OBJECT')
    except:
        pass
    return {'FINISHED'}

class ImportSomeData20(Operator, ImportHelper):
    """Import WWE 2k20 MDL object"""
    bl_idname = "import_test20.some_data"  # important since its how bpy.ops.import_test.some_data is constructed
    bl_label = "Import MDL!"

    # ImportHelper mixin class uses this
    filename_ext = ".mdl"

    filter_glob: StringProperty(
        default="*.mdl",
        options={'HIDDEN'},
        maxlen=255,  # Max internal buffer length, longer would be clamped.
    )

    use_false_bone_orientation: BoolProperty(
        name="[Debug] Use False Armature Orientation",
        description="Imports models using legacy rig import",
        default=False,
    )

    def execute(self, context):
        keywords = self.as_keywords(
            ignore=(
                "axis_forward",
                "axis_up",
                "global_scale",
                "check_existing",
                "filter_glob",
            ),
        )
        return read_some_data_20( context, self.filepath, keywords)

class ImportSomeData(Operator, ImportHelper):
    """Import WWE 2k22 MDL object"""
    bl_idname = "import_test.some_data"  # important since its how bpy.ops.import_test.some_data is constructed
    bl_label = "Import MDL!"

    # ImportHelper mixin class uses this
    filename_ext = ".mdl"

    filter_glob: StringProperty(
        default="*.mdl",
        options={'HIDDEN'},
        maxlen=255,  # Max internal buffer length, longer would be clamped.
    )

    use_false_bone_orientation: BoolProperty(
        name="[Debug] Use False Armature Orientation",
        description="Imports models using legacy rig import",
        default=False,
    )
    
    import_shape_keys: BoolProperty(
        name="Import Shape Keys",
        description="Imports all MDL Shape Key data.",
        default=False,
    )
    
    import_hpl: BoolProperty(
        name="Import HPL",
        description="Imports clipping masks for model",
        default=False,
    )

    def execute(self, context):
        keywords = self.as_keywords(
            ignore=(
                "axis_forward",
                "axis_up",
                "global_scale",
                "check_existing",
                "filter_glob",
            ),
        )
        return read_some_data( context, self.filepath, keywords )
        
bl_info = {
    "name": "WWE 2K22 (.MDL) Import/Export Add-On v1.16",
    "author": "Hanleys",
    "version": (1, 1, 6),
    "blender": (2, 93, 4),
    "description": "IO addon for importing/exporting WWE 2k22 .MDL models.",
    "category": "Object",
}

class ExportSomeData(Operator, ExportHelper):
    """Export selection as WWE 2k22 MDL object"""
    bl_idname = "export_test.some_data"  # important since its how bpy.ops.import_test.some_data is constructed
    bl_label = "Export MDL!"

    # ExportHelper mixin class uses this
    filename_ext = ".mdl"

    filter_glob: StringProperty(
        default="*.mdl",
        options={'HIDDEN'},
        maxlen=255,  # Max internal buffer length, longer would be clamped.
    )

    use_edgesplit: BoolProperty(
        name="Apply EdgeSplit / Preserve Flat Normals",
        description="Preserves flat shading of models. Alters topology.",
        default=False,
    )

    use_default_blender_normals: BoolProperty(
        name="Use Default/Internal Vertex Normals",
        description="Uses default internal vertex normals from model.",
        default=False,
    )

    export_blendshapes: BoolProperty(
        name="Export Shape Keys",
        description="Export with selected object's shape keys.",
        default=False,
    )

    def execute(self, context):
        keywords = self.as_keywords(
            ignore=(
                "axis_forward",
                "axis_up",
                "global_scale",
                "check_existing",
                "filter_glob",
            ),
        )

        return write_some_data(context, self.filepath, keywords)


# Only needed if you want to add into a dynamic menu
def menu_func_export(self, context):
    self.layout.operator(ExportSomeData.bl_idname, text="WWE2K22 MDL (.mdl)")
    
def menu_func_import(self, context):
    self.layout.operator(ImportSomeData.bl_idname, text="WWE2K22 MDL (.mdl)")

def menu_func_import_20(self, context):
    self.layout.operator(ImportSomeData20.bl_idname, text="WWE2K20 *Legacy MDL (.mdl)")


def register():
    bpy.utils.register_class(ExportSomeData)
    bpy.types.TOPBAR_MT_file_export.append(menu_func_export)
    bpy.utils.register_class(ImportSomeData)
    bpy.types.TOPBAR_MT_file_import.append(menu_func_import)
    bpy.utils.register_class(ImportSomeData20)
    bpy.types.TOPBAR_MT_file_import.append(menu_func_import_20)

def unregister():
    bpy.utils.unregister_class(ExportSomeData)
    bpy.types.TOPBAR_MT_file_export.remove(menu_func_export)
    bpy.utils.unregister_class(ImportSomeData)
    bpy.types.TOPBAR_MT_file_import.remove(menu_func_import)
    bpy.utils.unregister_class(ImportSomeData20)
    bpy.types.TOPBAR_MT_file_import.remove(menu_func_import_20)


if __name__ == "__main__":
    register()

    # test call
    bpy.ops.export_test.some_data('INVOKE_DEFAULT')
    bpy.ops.import_test.some_data('INVOKE_DEFAULT')
    bpy.ops.import_test20.some_data('INVOKE_DEFAULT')