import os
from typing import NamedTuple


def roundUp( numToRound , multiple ):
    if (multiple == 0): return numToRound
    remainder = numToRound % multiple
    if (remainder == 0): return numToRound
    return (numToRound + multiple - remainder)

def checkAlignment(  value ): 
    return roundUp( value , 4 )


class HPLObject(NamedTuple):
    sceneFlag: int
    name: str
    visibility_vector: list

class HPLLods(NamedTuple):
    lodLevel: int
    objects: list

def findDirHPL( modelPath ):
    #Search for first hpl
    dirs = modelPath.split("\\"); dirs.pop()
    path =  ("\\").join(dirs)
    for file in os.listdir(path):
        if file.endswith(".hpl"):
            return (("\\").join(dirs) + "\\" + file)
    return False

def getHPLPath( modelPath ):
    #Splits directory path and appends .hpl extension
    dirs = modelPath.split("\\")
    hplFile = ( dirs[-1].split(".mdl")[0] + ".hpl" )
    dirs.pop()
    return (("\\").join(dirs) + "\\" + hplFile) #return path


def readMESH(hplBuffer,hplBfDef,sceneFlagGlobal):
    #Get Visibility Flags
    meshCharLength = int.from_bytes( hplBuffer.read(4) , "little" )
    meshName = hplBuffer.read(meshCharLength).decode("utf-8")
    hplObj = HPLObject(sceneFlag=sceneFlagGlobal, name=meshName, visibility_vector=[])

    #align position
    alignPos = checkAlignment( hplBuffer.tell() )
    hplBuffer.seek( alignPos )
    numVisibleVerts = int.from_bytes( hplBuffer.read(4) , "little" )

    #Collect Array
    for x in range(numVisibleVerts):
        vIndex = int.from_bytes( hplBuffer.read(4) , "little" )
        hplObj.visibility_vector.append( vIndex )

    hplBfDef.objects.append(hplObj)
    return

def switch(hplBuffer,blockMagic,hplBfDef,sceneFlagGlobal):
    #Evaluate Type
    if blockMagic == b'LOD!':
        lodIdx = int.from_bytes( hplBuffer.read(4) , "little" )
        hplBfDef = HPLLods(lodLevel=lodIdx, objects=[])
    elif blockMagic == b'DATA':
        sceneFlagGlobal = int.from_bytes( hplBuffer.read(4) , "little" )
    elif blockMagic == b'MESH':
        readMESH(hplBuffer,hplBfDef,sceneFlagGlobal)
    return hplBfDef,sceneFlagGlobal

def parseHPL( hplBuffer ):
    #Create an HPL vector and store all parsed entries
    hplDefs = []
    fileSig = hplBuffer.read(4)
    if (fileSig != b'HPL!'): return False
    
    #Iterate/Collect data until file end
    bufferSig = hplBuffer.read(4)
    while( bufferSig != b'END!'):
        isRead=False; sceneFlagGlobal = 0 #Default as zero, modify as we parse
        if (not isRead):
            hplBfDef = HPLLods(lodLevel=None,objects=None); isRead = False
            subLODbf = b''
            while( subLODbf != b'LOD!' and subLODbf != b'END!'):
                #Parse data only through specified LOD level
                hplBfDef,sceneFlagGlobal = switch(hplBuffer,bufferSig,hplBfDef,sceneFlagGlobal)
                subLODbf = hplBuffer.read(4)
                bufferSig = subLODbf
            isRead=True; bufferSig = subLODbf
            hplDefs.append(hplBfDef)
    return hplDefs

def readHPLFile( hplPath ):
    if os.path.exists(hplPath):
        with open(hplPath, 'rb') as file:
            return parseHPL(file)  #interpret hpl
    else:
        return False

# if __name__ == "__main__":
#     mdlPath = "C:\\Users\\wauke\\Desktop\\jacket_10012.mdl"
#     if (".mdl" in mdlPath):
#         hplPath = getHPLPath(mdlPath)
#         hplRes = debugRunner(hplPath)
#         print( hplRes[0] )