from io_scene_mdl2022.mdl_utils_2022 import mdl_utils , scene_mdl_blender
from random import expovariate
from bpy import context
import bpy, bmesh
import configparser, os
import sys
from math import degrees
from math import radians
import mathutils
import math
from io_scene_mdl2022.mdl_utils_2022 import mbfd_dxgi_bin

global config
global config_path
global user_path
global conf_path
global hasConfig

config = configparser.ConfigParser()

user_path = bpy.utils.resource_path('USER')
conf_path = os.path.join(user_path, "config")
config_path = conf_path + "\mdl_skel.cache.ini" 

hasConfig = 'false'
use_edgesplit = 'false'
exportShapeKeys = 'false'

#global armature arrays
skelBones = []
skelBones_translate = []
skelBones_rotate = []
skelBones_parent = []

def exportArmature( armObj ):
    global skelBones
    global skelBones_translate
    global skelBones_rotate
    global skelBones_parent

    #set as active to access edit_bones
    bpy.context.view_layer.update()
    armature = armObj.data

    for bone in armature.bones:

        # Convert the edit bone's matrix to a 4x4 matrix
        skelBones.append( bone.name )
        boneMat = bone.matrix_local

        # Get the relative matrix by multiplying by inverse of parent matrix
        if bone.parent:
            parent_matrix = bone.parent.matrix_local
            boneMat  = parent_matrix.inverted() @ boneMat

        # Decompose for Pos , Rot , Scale vectors
        pos, rot, scale = boneMat.decompose()
        rot = rot.to_euler('XYZ')
                        
        skelBones_translate.append( pos )
        skelBones_rotate.append( rot )

        if bone.parent:
            skelBones_parent.append( bone.parent.name )
        else:
            skelBones_parent.append( 'NULL' )

    return 


def check_seams( model ):

    bpy.context.scene.tool_settings.use_uv_select_sync = True

    try:
        bpy.ops.object.mode_set(mode='EDIT')
        bpy.ops.mesh.select_mode(type="VERT")
        bpy.ops.mesh.select_all(action = 'SELECT')
    except:
        for screen in bpy.data.screens:
            for area in (a for a in screen.areas if a.type == 'VIEW_3D'):
                region = next((region for region in area.regions if region.type == 'WINDOW'), None)
                if region is not None:
                    override = {'area': area, 'region': region} 
                    bpy.ops.object.mode_set(mode='EDIT')
                    bpy.context.view_layer.objects.active = model

                    if bpy.context.active_object.mode != 'EDIT':
                        bpy.ops.object.mode_set(mode='EDIT', toggle=False)
                    else:
                        bpy.ops.object.mode_set(mode='OBJECT', toggle=False)
                        
                    bpy.context.scene.tool_settings.mesh_select_mode = (True, False, False)
                    for v in model.data.vertices:
                        v.select = True
        pass

    obj = model
    me = obj.data
    try:
        bm = bmesh.from_edit_mesh(me)
    except:
        bpy.ops.object.mode_set(mode='EDIT', toggle=False)
        bm = bmesh.from_edit_mesh(me)
        pass

    # old seams
    old_seams = [e for e in bm.edges if e.seam]
    # unmark
    for e in old_seams:
        e.seam = False

    # mark seams from uv islands
    try:
        bpy.ops.uv.seams_from_islands()
    except:
        for screen in bpy.data.screens:
            for area in (a for a in screen.areas if a.type == 'VIEW_3D'):
                region = next((region for region in area.regions if region.type == 'WINDOW'), None)
                if region is not None:
                    override = {'area': area, 'region': region} 
                    bpy.ops.uv.seams_from_islands()
        pass

    seams = [e for e in bm.edges if e.seam]
    # split on seams
    bmesh.ops.split_edges(bm, edges=seams)
    # re instate old seams.. could clear new seams.
    for e in old_seams:
        e.seam = True
    bmesh.update_edit_mesh(me)

    print("seams checked.")

    bpy.ops.object.mode_set(mode='OBJECT')

def init_config():
    global config
    global config_path
    global user_path
    global conf_path
    global hasConfig

    user_path = bpy.utils.resource_path('USER')
    conf_path = os.path.join(user_path, "config")
    config_path = conf_path + "\mdl_skel.cache.ini" 

    hasConfig = 'false'

    #clear old-cache
    if not os.path.exists(config_path):
        print("Config not found. Creating new...")
        write_file()
    else:
        # Read File
        hasConfig = 'true'
        config.read(config_path)
    
    print( "Config read" ) 

def write_file():
    config.write( open( config_path, 'w') )

def triangulate_object(obj):
    me = obj.data
    # Get a BMesh representation
    bm = bmesh.new()
    bm.from_mesh(me)

    bmesh.ops.triangulate(bm, faces=bm.faces[:])
    # V2.79 : bmesh.ops.triangulate(bm, faces=bm.faces[:], quad_method=0, ngon_method=0)

    # Finish up, write the bmesh back to the mesh
    bm.to_mesh(me)
    bm.free()
    
def triangulate_edit_object(obj):
    me = obj.data
    # Get a BMesh representation
    bm = bmesh.from_edit_mesh(me)

    bmesh.ops.triangulate(bm, faces=bm.faces[:])
    # V2.79 : bmesh.ops.triangulate(bm, faces=bm.faces[:], quad_method=0, ngon_method=0)

    # Show the updates in the viewport
    # and recalculate n-gon tessellation.
    bmesh.update_edit_mesh(me, True)


def getActiveModels():

    selection_names = bpy.context.selected_objects
    selection_ = []
    
    for sel in selection_names:
        if sel.type == 'MESH':
            selection_.append( sel )
            
    return selection_

def getFaceList( model ):
    isQuad = 'false'
    
    faceList = []
    
    for f in model.data.polygons:
        faceTuple = []
        for idx in f.vertices:
            faceTuple.append( model.data.vertices[idx].index + 1)
                
        if ( len( faceTuple ) != 3 ):
            isQuad = 'true'
        
        faceList.append(  faceTuple )
        
    return isQuad , faceList

##  VColor Conversion -- Code by Grix    
def to_lin(s):
    if s <= 0.0404482362771082:
        lin = s / 12.92
    else:
        lin = pow(((s + 0.055) / 1.055), 2.4)
    return lin

def checkModelFlags( modelID ):
    if ( "!" in modelID):
        return "ENT_ONLY"
    if ( "$" in modelID):
        return "RING_ONLY"
    if ( modelID.find('$')==-1 and 
         modelID.find('!')==-1):
         return "false"

def parseModels( sceneModels  , EXPORT_EDGESPLIT ):
    
    meshArr = []
    vertArr = []
    faceArr = []
    normArr = []
    uvArr = []
    skin = []
    meshNames = []
    specialFlags = []
    vColor_sets = []
    scene_models = []

    for model in sceneModels:

        check_seams( model )
        
        print( "\nModel " + model.name )
        bpy.ops.object.mode_set(mode='OBJECT')

        modelname = model.name

        specialFlags.append( checkModelFlags( modelname ) )
        modelname = modelname.replace( "!","")
        modelname = modelname.replace( "$","")

        if ( "." in  modelname ):
            mdlstr = modelname.split(".")
            meshArr.append( mdlstr[0] )
        else:
            meshArr.append( modelname )

        model_meshID = model.data.name
        
        if ( "." in model_meshID ):
            mdlstr = model_meshID.split(".")
            meshNames.append( mdlstr[0] )
        else:
            meshNames.append( model_meshID )
            
        ##get faces
        isQuad = 'false'
        faceInds = getFaceList( model )

        if ( faceInds[0] == 'true' ):
            print ("is Quad")
            triangulate_object( model )

            ######## AutoGen EdgeSplit Data ########
            model.data.calc_normals_split()

            #bool checkPreModifiers
            hasEdgeSplit = 'false'
            for modifier in model.modifiers:
                if modifier.type == "EDGE_SPLIT":
                    hasEdgeSplit = 'true'

            if EXPORT_EDGESPLIT == True:
                hasEdgeSplit = 'false'
            else:
                hasEdgeSplit = 'true'

            if hasEdgeSplit == 'false':
                edgeSplitModif = model.modifiers.new(name = 'OptimNorms' , type='EDGE_SPLIT')

            bpy.ops.object.modifier_apply(modifier="OptimNorms")

            ####################################
            
            faceInds = getFaceList( model )

        else:
            ######## AutoGen EdgeSplit Data ########
            model.data.calc_normals_split()

            #bool checkPreModifiers
            hasEdgeSplit = 'false'
            for modifier in model.modifiers:
                if modifier.type == "EDGE_SPLIT":
                    hasEdgeSplit = 'true'
                                
            if EXPORT_EDGESPLIT == True:
                hasEdgeSplit = 'false'
            else:
                hasEdgeSplit = 'true'
                        
            if hasEdgeSplit == 'false':
                edgeSplitModif = model.modifiers.new(name = 'OptimNorms' , type='EDGE_SPLIT')

            bpy.ops.object.modifier_apply(modifier="OptimNorms")

            ####################################

            faceInds = getFaceList( model )      

        faceArr.append( faceInds )
        model.data.calc_normals_split()

        ##get verts
        verts = [model.matrix_world @ vert.co for vert in model.data.vertices]
        plain_verts = [vert.to_tuple() for vert in verts]
        
        ##get norms

        modelNorms = list( range( len( model.data.vertices ) ) )
        modelNorms = [ ( 0.0 , 0.0 , 1.0 ) for defaultNorm in modelNorms ] #default norm

        for loop in model.data.loops:

            loopNormal = loop.normal
            loop_vIdx = loop.vertex_index
            
            normal_local = loopNormal.to_4d()
            normal_local.w = 0
            normal_local = (model.matrix_world @ normal_local).to_3d()

            modelNorms[ loop_vIdx ] = normal_local.to_tuple()

        # print("nroms:")
        # print( modelNorms ) 

        vertArr.append( plain_verts )
        normArr.append( modelNorms )

        ##getUVs
        uv_sets = model.data.uv_layers
        uv_layers = list( range( len( uv_sets) ) )
        print( "Model has UV Layers: " , len( uv_sets ) )

        # Loops per UV Layer
        uv_layer_idx = 0
        for uv_layer in uv_sets:
            model_uvs = list( range( len( plain_verts ) ) )

            # Loops per face
            for face in model.data.polygons:
                for vert_idx, loop_idx in zip(face.vertices, face.loop_indices):
                    uv_coords = uv_layer.data[loop_idx].uv
                    model_uvs[vert_idx] = ( uv_coords.x , uv_coords.y )

            uv_layers[ uv_layer_idx ] = model_uvs
            uv_layer_idx += 1

        uvArr.append( uv_layers )

        ##getVColors
        hasColor = 'true'
        model_vcolors = []
        try:
            color_layer = model.data.vertex_colors["Col"]
            model_vcolors = list(range( len(plain_verts) ))
            i = 0
            for poly in model.data.polygons:
                for v_ix, l_ix in zip(poly.vertices, poly.loop_indices):
                    vcolors = color_layer.data[i].color
                    model_vcolors[v_ix] = ( to_lin(vcolors[0]), to_lin(vcolors[1]) , to_lin(vcolors[2]) , vcolors[3] )
                    i += 1
            print( "hasColor" )
        except:
            hasColor = 'false'
            pass
        
        vColor_sets.append( model_vcolors )

        ##skin
        model_skin_set = []
        print( model.parent )
        
        if type(model.parent) != type(None):
            
            # limitBoneWeights to 4
            bpy.context.view_layer.objects.active  = model
            bpy.ops.object.vertex_group_limit_total(group_select_mode='ALL', limit=4)
           
            ##getBnes
            if( len(skelBones) == 0 ):
                if ( model.parent ).type == 'ARMATURE':
                    exportArmature( model.parent )

            ##accessVGroups
            objGroups = []
            for group in model.vertex_groups:
                objGroups.append( group.name )
            
            mesh = model.data
            skinInd = []
            skinWeights = []
             
            #collectSkinInfo
            for v in mesh.vertices:
                groupTuple = []
                indTuple = []
                
                #collectBones
                for elem in v.groups:
                    indTuple.append( objGroups[ elem.group ] )
                
                #collectWeights
                for elem in v.groups:
                    groupTuple.append( elem.weight )
                    
                skinInd.append( indTuple )
                skinWeights.append( groupTuple )
            
            model_skin_set.append( skinInd )
            model_skin_set.append( skinWeights )
            
            scene_models.append( model )
            skin.append( model_skin_set )
                    
        else:
            print( "model has no armature" )
    
    return meshArr , vertArr , faceArr , normArr , uvArr , skelBones , skelBones_translate , skelBones_rotate , skin , skelBones_parent , meshNames, specialFlags , vColor_sets , scene_models
##           0         1        2         3         4         5         6                      7                8          9               10           11             12           13

def getDeltaMatrixValue( deltaGroup ):

    deltaLookupTable = { '0': [],    
    '1': [0], '2': [1], '3': [0,1], '4': [2],
    '5': [0,2], '6': [1,2], '7': [0,1,2], '8': [3],
    '9': [0,3], 'A': [1,3], 'B': [0,1,3], 'C': [2,3],
    'D': [0,2,3], 'E': [1,2,3], 'F': [0,1,2,3]
    }

    deltaValue = [k for k, v in deltaLookupTable.items() if deltaGroup == v][0]

    return deltaValue

import statistics

def getMinMaxValues( vert_deltas ):

    lowest_value_x = min(vert_deltas[0])
    lowest_value_y = min(vert_deltas[1])
    lowest_value_z = min(vert_deltas[2])

    highest_value_x = max(vert_deltas[0])
    highest_value_y = max(vert_deltas[1])
    highest_value_z = max(vert_deltas[2])

    index_x = vert_deltas[0].index( lowest_value_x )
    index_y = vert_deltas[1].index( lowest_value_y )
    index_z = vert_deltas[2].index( lowest_value_z )

    index_x_h = vert_deltas[0].index( highest_value_x )
    index_y_h = vert_deltas[1].index( highest_value_y )
    index_z_h = vert_deltas[2].index( highest_value_z )

    return lowest_value_x , lowest_value_y , lowest_value_z , index_x , index_y , index_z , highest_value_x , highest_value_y , highest_value_z , index_x_h , index_y_h , index_z_h

def getWeightData_PRE( vert_arr , modif_arr , average_minimum_x , average_minimum_y , average_minimum_z ):
    default_verts = vert_arr
    modif_verts = modif_arr

    verts_x = []
    mverts_x = []
    vert_delta_x = []

    for x in range( len(default_verts) ):
        vert_x  = default_verts[x].co.x

        mvert_x = modif_verts[x].co.x

        delta_x = mvert_x - vert_x

        vert_delta_x.append( delta_x )
        verts_x.append( vert_x )
        mverts_x.append( mvert_x )


    lowest_x = max( vert_delta_x )
    idx = vert_delta_x.index( lowest_x )
    reg_x = verts_x[idx]
    mod_x = mverts_x[idx]

    return [ lowest_x , reg_x , mod_x ]

def getShapeData_MAIN( inputModel , blendShape , arraySize , floor ):

    default_verts = []
    modif_verts = []
    modif_indices = []

    vert_deltas = []
    vert_deltas_x = []
    vert_deltas_y = []
    vert_deltas_z = []

    for vert in inputModel.data.vertices:
        default_verts.append( vert )
        new_vert = blendShape.data[vert.index]
        modif_verts.append( new_vert )
    
    for vert in default_verts:
        v_Idx = vert.index
        modif_vert = modif_verts[v_Idx]

        #analyze vertices
        orig_x = vert.co.x
        orig_y = vert.co.y
        orig_z = vert.co.z

        modif_x = modif_vert.co.x
        modif_y = modif_vert.co.y
        modif_z = modif_vert.co.z

        vert_deltas_x.append( modif_x - orig_x)
        vert_deltas_y.append( modif_y - orig_y)
        vert_deltas_z.append( modif_z - orig_z)
        

        if [ orig_x , orig_y , orig_z ] != [ modif_x , modif_y , modif_z ]:
            modif_indices.append( v_Idx )
    vert_deltas = [ vert_deltas_x , vert_deltas_y , vert_deltas_z ]
    

    #group modified indices in array of 4
    all = list(range(len(default_verts)))
    n = 4
    v_groups = [all[k:k+n] for k in range(0, len(all), n)]
    delta_groups = []

    
    for group in v_groups:
        out_group = []
        for idx in range( len(group)):
            vert = group[idx]
            if (len(modif_indices)!=0):
                if vert == modif_indices[0]:
                    out_group.append( idx )
                    modif_indices.pop(0)
        delta_groups.append( out_group )

    #getArray value
    outArr = []
    for group in delta_groups:
        bitValue = getDeltaMatrixValue( group )
        outArr.append( bitValue )
    
    while len(outArr) % 8:
        outArr.append('0')

    #binary_output -indexBuffer
    indexBuffer = io.BytesIO(b'')

    #swap nibbles
    s=len(outArr)
    if s%2!=0:
        s=s-1
    for i in range(0,s,2):
        outArr[i],outArr[i+1]=outArr[i+1],outArr[i]


    #append data
    alignBit = 0
    nibbleBit = 0
    DWORD_MatrixSize = 0
    high_bit = 0

    # print( "Length: " , len(outArr) )
    for bit in outArr:

        bitValue = int(bit,16)
        bitSize = len( mdl_utils.DELTA_ELEMENTS_FROM_BIT( bit , 0) )
        DWORD_MatrixSize += bitSize

        nibbleBit+=1
        alignBit+=1

        if (nibbleBit > 1):
            bitValue += ( high_bit << 4 )
            indexBuffer.seek( 0 , 2 );indexBuffer.write(  int(bitValue).to_bytes(1, 'little')  )

            nibbleBit = 0
            high_bit = 0

        if (alignBit > 7):
            indexBuffer.seek( 0 , 2 );indexBuffer.write(  int(arraySize+floor).to_bytes(4, 'little')  )

            floor+=((DWORD_MatrixSize*8)/4) 
            alignBit = 0
            high_bit = 0
            nibbleBit = 0
            DWORD_MatrixSize = 0

        high_bit = bitValue

    indexBuffer.seek(0)
    return indexBuffer.read() , floor , default_verts , modif_verts , vert_deltas


def formatTEXT( sceneData ): 
    
    mainText = []
    
    #addBoneNames
    for item in sceneData[5]:
        mainText.append( item )
        
    #addMeshIDs
    for item in sceneData[0]:
        mainText.append( item )
        
    
    #addDataTypes
    mainText.append( 'POSITION' )
    mainText.append( 'R32_G32_B32' )
    mainText.append( 'float' )
    mainText.append( 'NORMAL' )
    mainText.append( 'R8_G8_B8_A8' )
    mainText.append( 'snorm' )
    mainText.append( 'TANGENT' )
    mainText.append( 'BINORMAL' )
    mainText.append( 'R8' )
    mainText.append( 'COLOR' )
    mainText.append( 'unorm' )
    mainText.append( 'TEXCOORD' )
    mainText.append( 'R32_G32' )
    mainText.append( 'R16_G16_B16_A16' )
    mainText.append( 'BLENDINDICES' )
    mainText.append( 'uint' )
    mainText.append( 'BLENDWEIGHTS' )
    mainText.append( 'R32_G32_B32_A32' )
    mainText.append( '' )
    
    #append shape-key ID's
    for item in sceneData[13]:

        num_of_SK_blocks = 0
        if item.data.shape_keys:
            num_of_SK_blocks = len( item.data.shape_keys.key_blocks )

        if num_of_SK_blocks != 0:
            for shape in item.data.shape_keys.key_blocks:
                shape_ID = shape.name
                
                if shape_ID != "Basis":
                    if shape not in mainText:
                        mainText.append( shape_ID )

    #addMeshNames
    for i in range( len( sceneData[0] ) ):
        mainText.append( sceneData[10][i] )
    
    
    return mainText


import io

def get_binaryTEXT( mainTEXT ):
    
    textCount = len( mainTEXT )
    in_memory_sub = io.BytesIO(b'')

    in_memory_sub.seek( 0 , 2 );in_memory_sub.write( textCount.to_bytes(4, 'little') )
    
    strOffset = 0
    for a in range( len(mainTEXT) ):
        itemString = mainTEXT[a]
        in_memory_sub.seek( 0 , 2 )
        in_memory_sub.write( strOffset.to_bytes(4, 'little') )
        strOffset += len(itemString)+1
    
    for a in range( len(mainTEXT) ):
        itemString = mainTEXT[a]
        in_memory_sub.seek( 0 , 2 );in_memory_sub.write( itemString.encode('utf-8') )
        in_memory_sub.seek( 0 , 2 );in_memory_sub.write( b'\x00' )
    
    totalSize = 8 + len( in_memory_sub.getvalue() )
    print( totalSize )
    
    while ( totalSize % 8 ):
        in_memory_sub.seek( 0 , 2 );in_memory_sub.write( b'\x00' )
        totalSize += 1
    
    in_memory = io.BytesIO(b'TEXT')
    in_memory.seek( 0 , 2 );in_memory.write( (totalSize-8).to_bytes(4, 'little') )
 
    
    return in_memory.getvalue() , in_memory_sub.getvalue()
    
def get_binaryMDLStart( ):
    
    in_memory = io.BytesIO(b'MDL!')
    in_memory.seek( 0 , 2 );in_memory.write(  int(32).to_bytes(4, 'little')  )
    in_memory.seek( 0 , 2 );in_memory.write(  b'\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF'  )
    in_memory.seek( 0 , 2 );in_memory.write(  b'\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF'  )
    in_memory.seek( 0 , 2 );in_memory.write(  b'\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF'  )


    return in_memory.getvalue()

import ast
import struct
import binascii

def float_to_hex(f):
    return struct.pack('<f', f)

def get_binaryBONE( mainText , skelBones , skelTrans , skelRot , skelPar ):
    
    in_memory = io.BytesIO(b'BONE')
    
    boneCount = len( skelBones )
    in_memory_sub = io.BytesIO(b'')
    in_memory_sub.seek( 0 , 2 );in_memory_sub.write( boneCount.to_bytes(4, 'little') )
    
    for a in range( len(skelBones) ):
        bone_name = skelBones[a]
        bone_index = mainText.index( bone_name )
        in_memory_sub.seek( 0 , 2 );in_memory_sub.write( bone_index.to_bytes(2, 'little') )
        
        bone_parent = skelPar[a]
        if ( bone_parent != 'NULL' ):
            bone_parent_index = mainText.index( bone_parent )
            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( bone_parent_index.to_bytes(2, 'little') )
        else:
            bone_parent_index = 65535
            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( bone_parent_index.to_bytes(2, 'little') )

        bone_translate = skelTrans[a]
        bone_rot =  skelRot[a]
        bone_tr_hex = [float_to_hex(i) for i in bone_translate]
        bone_ro_hex = [ float_to_hex(i) for i in bone_rot]
        
        for c in range(len(bone_tr_hex)):
            in_memory_sub.seek( 0 , 2 )
            in_memory_sub.write( bone_tr_hex[c])
        for c in range(len(bone_ro_hex)):
            in_memory_sub.seek( 0 , 2 )
            in_memory_sub.write( bone_ro_hex[c])
    
    in_memory = io.BytesIO(b'BONE')
    in_memory.seek( 0 , 2 );in_memory.write( (len(in_memory_sub.getvalue())).to_bytes(4, 'little') )       
    return in_memory.getvalue() , in_memory_sub.getvalue()
    
def get_binaryMTL( sceneData , mainText ):
    
    in_memory = io.BytesIO(b'MTL!')
    in_memory_sub = io.BytesIO(b'')
    in_memory_sub.seek( 0 , 2 );in_memory_sub.write( (len(sceneData[0])).to_bytes(4, 'little') )
    
    for i in range( len(sceneData[0]) ):
        meshId = sceneData[0][i]
        meshIndex = mainText.index( meshId )
        in_memory_sub.seek( 0 , 2 );in_memory_sub.write( meshIndex.to_bytes(4, 'little') )
    
    in_memory = io.BytesIO(b'MTL!')
    in_memory.seek( 0 , 2 );in_memory.write( (len(in_memory_sub.getvalue())).to_bytes(4, 'little') )    
    
    return in_memory.getvalue() , in_memory_sub.getvalue()


def get_binaryMBFD( sceneData , mainText , hasRig , EXPORT_SHAPE_KEYS ):
    in_memory_sub = io.BytesIO(b'')
    in_memory_sub.seek( 0 , 2 );in_memory_sub.write( (len(sceneData[0])).to_bytes(4, 'little') )
    skinSetIndex = 0
    
    for i in range( len(sceneData[0]) ):
        meshString = sceneData[10][i]
        meshIndex = mainText.index( meshString )
        blockCount = 0

        if ( hasRig == 'true'):
            blockCount = 7
        else:
            blockCount = 5

        if len( sceneData[12][i] ) != 0:
            blockCount += 1

        model_uv_layers = len( sceneData[4][i] )
        blockCount += ( model_uv_layers - 1 )

        meshVerts = sceneData[ 1 ][i]
        specialFlag = sceneData[11][i]

        in_memory_sub.seek( 0 , 2 );in_memory_sub.write( meshIndex.to_bytes(4, 'little') )
        if ( specialFlag == "ENT_ONLY" ):
            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( int(129).to_bytes(4, 'little') ) #flag 1
            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( int(0).to_bytes(2, 'little') ) #flag 2
        if ( specialFlag == "RING_ONLY" ):
            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( int(130).to_bytes(4, 'little') ) #flag 1
            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( int(0).to_bytes(2, 'little') ) #flag 2
        if ( specialFlag == "false" ):
            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( int(128).to_bytes(4, 'little') ) #flag 1
            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( int(0).to_bytes(2, 'little') ) #flag 2
        in_memory_sub.seek( 0 , 2 );in_memory_sub.write( b'\xFF\xFF\xFF\xFF' ) #mshhash
        in_memory_sub.seek( 0 , 2 );in_memory_sub.write( b'\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF' ) #mshhash
        in_memory_sub.seek( 0 , 2 );in_memory_sub.write( b'\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF' ) #mshhash
        in_memory_sub.seek( 0 , 2 );in_memory_sub.write( b'\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF' ) #mshhash
        ####VERTICES
        
        posIndex = mainText.index( 'POSITION' )
        dxgbIndex = mainText.index( 'R32_G32_B32' )
        floatIndex = mainText.index( 'float' )
        
        in_memory_sub.seek( 0 , 2 );in_memory_sub.write( len(meshVerts).to_bytes(4, 'little') ) #meshVerts
        in_memory_sub.seek( 0 , 2 );in_memory_sub.write( int(blockCount).to_bytes(4, 'little') ) #meshBlocks --
        in_memory_sub.seek( 0 , 2 );in_memory_sub.write( b'\xA4\xCE\x1A\x5A\x67\x7D\x20\x25\x5F\x8E\xB5\x47' ) #vertHash
        in_memory_sub.seek( 0 , 2 );in_memory_sub.write( posIndex.to_bytes(2, 'little') ) #vertHash
        in_memory_sub.seek( 0 , 2 );in_memory_sub.write( dxgbIndex.to_bytes(2, 'little') ) #vertHash
        in_memory_sub.seek( 0 , 2 );in_memory_sub.write( floatIndex.to_bytes(4, 'little') ) #vertHash
        in_memory_sub.seek( 0 , 2 );in_memory_sub.write( int(0).to_bytes(2, 'little') ) #align
        
        for a in range( len(meshVerts) ):
            binFloat_x = 1.0 * meshVerts[a][0]
            binFloat_y = -1.0 * meshVerts[a][2]
            binFloat_z = 1.0 * meshVerts[a][1]
            binFloat_x = float_to_hex( binFloat_x )
            binFloat_y = float_to_hex( binFloat_y )
            binFloat_z = float_to_hex( binFloat_z )
            
            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( binFloat_x ) #meshVertsx
            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( binFloat_y ) #meshVertsy
            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( binFloat_z ) #meshVertsz
        
        
        ####NORMALS
        
        meshNorms = sceneData[ 3 ][i]
        
        nrmIndex = mainText.index( 'NORMAL' )
        dxgbIndex = mainText.index( 'R32_G32_B32' )
        floatIndex = mainText.index( 'float' )
        
        in_memory_sub.seek( 0 , 2 );in_memory_sub.write( b'\x14\x2F\x99\x24\x67\x7D\x20\x25\x5F\x8E\xB5\x47' ) #nrmlHash
        in_memory_sub.seek( 0 , 2 );in_memory_sub.write( nrmIndex.to_bytes(2, 'little') ) 
        in_memory_sub.seek( 0 , 2 );in_memory_sub.write( dxgbIndex.to_bytes(2, 'little') ) 
        in_memory_sub.seek( 0 , 2 );in_memory_sub.write( floatIndex.to_bytes(4, 'little') ) 
        
        for a in range( len(meshNorms) ):
            binFloat_x = 1.0 * meshNorms[a][0]
            binFloat_y = 1.0 * meshNorms[a][1]
            binFloat_z = -1.0 * meshNorms[a][2] ##checkths
            binFloat_x = float_to_hex( binFloat_x )
            binFloat_y = float_to_hex( binFloat_y )
            binFloat_z = float_to_hex( binFloat_z )
                
            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( binFloat_x ) #meshVertsx
            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( binFloat_z ) #meshVertsy
            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( binFloat_y ) #meshVertsz
            
        ####TANGENT
        
        nrmIndex = mainText.index( 'TANGENT' )
        dxgbIndex = mainText.index( 'R8_G8_B8_A8' )
        floatIndex = mainText.index( 'snorm' )
        
        in_memory_sub.seek( 0 , 2 );in_memory_sub.write( b'\xB3\x88\x28\x62\xE0\x77\x7C\x76\xAA\x58\x6A\xCC' ) #nrmlHash
        in_memory_sub.seek( 0 , 2 );in_memory_sub.write( nrmIndex.to_bytes(2, 'little') )
        in_memory_sub.seek( 0 , 2 );in_memory_sub.write( dxgbIndex.to_bytes(2, 'little') )
        in_memory_sub.seek( 0 , 2 );in_memory_sub.write( floatIndex.to_bytes(4, 'little') )
        
        for a in range( len(meshVerts) ):
            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( int(255).to_bytes(4, 'little') )
        
        ####BINORMALS
        
        nrmIndex = mainText.index( 'BINORMAL' )
        dxgbIndex = mainText.index( 'R8' )
        floatIndex = mainText.index( 'snorm' )
        
        in_memory_sub.seek( 0 , 2 );in_memory_sub.write( b'\xAD\xE0\x21\x1B\xC5\x82\x47\xD3\xAA\x58\x6A\xCC' ) #nrmlHash
        in_memory_sub.seek( 0 , 2 );in_memory_sub.write( nrmIndex.to_bytes(2, 'little') )
        in_memory_sub.seek( 0 , 2 );in_memory_sub.write( dxgbIndex.to_bytes(2, 'little') )
        in_memory_sub.seek( 0 , 2 );in_memory_sub.write( floatIndex.to_bytes(4, 'little') )
        
        for a in range( len(meshVerts) ):
            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( int(255).to_bytes(1, 'little') )    
        
        ##align    
        while ( len(in_memory_sub.getvalue() ) % 4 ):
            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( int(0).to_bytes(1, 'little') )    
        
        ####VCOLORS
        mdl_colors = sceneData[12][i]

        if ( len (mdl_colors) ) != 0:
            nrmIndex = mainText.index( 'COLOR' )
            dxgbIndex = mainText.index( 'R8_G8_B8_A8' )
            floatIndex = mainText.index( 'unorm' )

            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( b'\x6B\x63\x97\xC5\xE0\x77\x7C\x76\x42\x39\x0D\x5C' ) #nrmlHash
            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( nrmIndex.to_bytes(2, 'little') )
            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( dxgbIndex.to_bytes(2, 'little') )
            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( floatIndex.to_bytes(4, 'little') )

            for a in range( len(meshVerts) ):
                mesh_vcol = mdl_colors[a]

                in_memory_sub.seek( 0 , 2 );in_memory_sub.write( int(mesh_vcol[0] * 255.0 ).to_bytes(1, 'little') )  
                in_memory_sub.seek( 0 , 2 );in_memory_sub.write( int(mesh_vcol[1] * 255.0 ).to_bytes(1, 'little') )  
                in_memory_sub.seek( 0 , 2 );in_memory_sub.write( int(mesh_vcol[2] * 255.0 ).to_bytes(1, 'little') ) 
                in_memory_sub.seek( 0 , 2 );in_memory_sub.write( int(mesh_vcol[3] * 255.0 ).to_bytes(1, 'little') )  

        ####TEXCOORDS
        
        for uv_layer_idx in range( model_uv_layers ):

            meshUVs = sceneData[4][i][uv_layer_idx]
            nrmIndex = mainText.index( 'TEXCOORD' )
            dxgbIndex = mainText.index( 'R32_G32' )
            floatIndex = mainText.index( 'float' )
            
            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( b'\x59\xA1\x7B\x3C\x27\x02\x45\x7F\x5F\x8E\xB5\x47' ) #nrmlHash
            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( nrmIndex.to_bytes(2, 'little') )
            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( dxgbIndex.to_bytes(2, 'little') )
            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( floatIndex.to_bytes(4, 'little') )
            
            for a in range( len(meshUVs) ):
                try:
                    binFloat_uvx = 1.0 * meshUVs[a][0]
                except:
                    pass
                    binFloat_uvx = 1.0
                try:
                    binFloat_uvy = -1.0 * (-1.0 + meshUVs[a][1])
                except:
                    pass
                    binFloat_uvy = 1.0
                binFloat_uvx = float_to_hex( binFloat_uvx )
                binFloat_uvy = float_to_hex( binFloat_uvy )
                
                in_memory_sub.seek( 0 , 2 );in_memory_sub.write( binFloat_uvx ) #meshVertsx
                in_memory_sub.seek( 0 , 2 );in_memory_sub.write( binFloat_uvy ) #meshVertsx


        ####Skin
        mdl_skin = sceneData[8]
        
        if ( len (mdl_skin) ) != 0:
            
            if not i > ( len (mdl_skin)  - 1 ):
                model_skin_set = mdl_skin[i]
                
                ####BLENDINDICES
                nrmIndex = mainText.index( 'BLENDINDICES' )
                dxgbIndex = mainText.index( 'R16_G16_B16_A16' )
                floatIndex = mainText.index( 'uint' )
                
                in_memory_sub.seek( 0 , 2 );in_memory_sub.write( b'\xB3\x6E\x8F\x6B\x4C\x21\x26\x64\x35\x89\x93\x02' ) #bonehash
                in_memory_sub.seek( 0 , 2 );in_memory_sub.write( nrmIndex.to_bytes(2, 'little') )
                in_memory_sub.seek( 0 , 2 );in_memory_sub.write( dxgbIndex.to_bytes(2, 'little') )
                in_memory_sub.seek( 0 , 2 );in_memory_sub.write( floatIndex.to_bytes(4, 'little') )
                
                for a in range( len(model_skin_set[0]) ):
               
                    for b in range( 4 ):

                        if ( b > len( model_skin_set[0][a] ) - 1 ):
                            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( int(0).to_bytes(2, 'little') )
                        else:
                            boneName = model_skin_set[0][a][b]
                            boneIndex = mainText.index( boneName )
                        
                            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( boneIndex.to_bytes(2, 'little') )
                        
                ###BLENDWEIGHTS
                nrmIndex = mainText.index( 'BLENDWEIGHTS' )
                dxgbIndex = mainText.index( 'R8_G8_B8_A8' )
                floatIndex = mainText.index( 'unorm' )
                
                in_memory_sub.seek( 0 , 2 );in_memory_sub.write( b'\x9B\x1A\x8F\x49\xE0\x77\x7C\x76\x42\x39\x0D\x5C' ) #bonehash
                in_memory_sub.seek( 0 , 2 );in_memory_sub.write( nrmIndex.to_bytes(2, 'little') )
                in_memory_sub.seek( 0 , 2 );in_memory_sub.write( dxgbIndex.to_bytes(2, 'little') )
                in_memory_sub.seek( 0 , 2 );in_memory_sub.write( floatIndex.to_bytes(4, 'little') )   
                
                for a in range( len(model_skin_set[1]) ):
                    for b in range( 4 ):
                        if ( b > len( model_skin_set[1][a] ) - 1 ):
                            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( int(0).to_bytes(1, 'little') )
                        else:
                            boneWeight = model_skin_set[1][a][b]
                            boneWeight = boneWeight * 255
                            
                            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( int(boneWeight).to_bytes(1, 'little') )
        else:
            print( "model is not skinned")              
        
        ##Vertex Remapping Section (null)
        for a in range( len(meshVerts) ):
            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( b'\x00\x00\x00\x00' ) #meshVertsx

        #add Shape Data
        objModel = sceneData[13][i]

        #check if contains morphs
        morphCount = 0
        if objModel.data.shape_keys:
            morphCount = len( objModel.data.shape_keys.key_blocks )

        #toggle shape keys
        if EXPORT_SHAPE_KEYS == False:
            morphCount = 0

        if morphCount != 0:

            for shape in objModel.data.shape_keys.key_blocks:
                shape_ID = shape.name

                if shape_ID =="Basis":
                    morphCount -= 1 

            #write
            blankEntry = mainText.index( '' )
            set_count = mbfd_dxgi_bin.roundUp( ( (len(objModel.data.vertices)) / 32 ) , 1 )
            # print( "SHAPE COUNT: " ,  set_count , morphCount )

            arraySize = ( set_count * 2 ) * morphCount
            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( int(morphCount).to_bytes(4, 'little') ) #morph Count

            #default arrs
            shapeVertArr = []
            shapeModifArr = []
            shapeDeltas = []

            #ptr buffer
            floor = 0

            #tfm vars min
            lowest_x = []
            lowest_y = []
            lowest_z = []

            indices_x = []
            indices_y = []
            indices_z = []

            #tfm vars max
            highest_x = []
            highest_y = []
            highest_z = []

            indices_x_h = []
            indices_y_h= []
            indices_z_h = []


            #wt vars
            precision_x = []
            precision_y = []
            precision_z = []

            ##index+transpose data
            for shape in objModel.data.shape_keys.key_blocks:
                shape_ID = shape.name

                if shape_ID != "Basis":
                    in_memory_sub.seek( 0 , 2 );in_memory_sub.write(  int(mainText.index(shape_ID)).to_bytes(2, 'little') ) 
                
            ##align    
            while ( len(in_memory_sub.getvalue() ) % 4 ):
                in_memory_sub.seek( 0 , 2 );in_memory_sub.write( int(0).to_bytes(1, 'little') )
            
            #vertices
            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( int(len( objModel.data.vertices )).to_bytes(4, 'little') )
            
            ##index+transpose data
            dataBuffer = io.BytesIO(b'')

            for shape in objModel.data.shape_keys.key_blocks:
                shape_ID = shape.name
                
                if shape_ID != "Basis":
                    shapeData , floor , default_verts , modif_verts , vert_deltas = getShapeData_MAIN( objModel , shape , arraySize , floor )
                    lowest_value_x , lowest_value_y , lowest_value_z , \
                    index_x , index_y , index_z , \
                    highest_value_x , highest_value_y , highest_value_z , \
                    index_x_h , index_y_h , index_z_h = getMinMaxValues( vert_deltas )

                    lowest_x.append( lowest_value_x )
                    lowest_y.append( lowest_value_y )
                    lowest_z.append( lowest_value_z )

                    highest_x.append( highest_value_x )
                    highest_y.append( highest_value_y )
                    highest_z.append( highest_value_z )

                    indices_x.append(  [ lowest_value_x , index_x , default_verts[index_x] , modif_verts[index_x] ] )
                    indices_y.append(  [ lowest_value_y , index_y , default_verts[index_y] , modif_verts[index_y] ] )
                    indices_z.append(  [ lowest_value_z , index_z , default_verts[index_z] , modif_verts[index_z] ] )

                    indices_x_h.append(  [ highest_value_x , index_x_h , default_verts[index_x_h] , modif_verts[index_x_h] ] )
                    indices_y_h.append(  [ highest_value_y , index_y_h , default_verts[index_y_h] , modif_verts[index_y_h] ] )
                    indices_z_h.append(  [ highest_value_z , index_z_h , default_verts[index_z_h] , modif_verts[index_z_h] ] )
                    
                    #append-cache
                    shapeVertArr.append( default_verts )
                    shapeModifArr.append( modif_verts )
                    shapeDeltas.append( vert_deltas )

                    dataBuffer.write( shapeData )

            
            #get average transpose
            average_minimum_x = min(lowest_x)
            average_minimum_y = min(lowest_y)
            average_minimum_z = min(lowest_z)

            highx_index = indices_x_h[ highest_x.index( max(highest_x)) ]
            highy_index = indices_y_h[ highest_y.index( max(highest_y)) ]
            highz_index = indices_z_h[ highest_z.index( max(highest_z)) ]
            
            src_vert_x_h = highx_index[2].co.x
            mdf_vert_x_h = highx_index[3].co.x

            src_vert_y_h = highy_index[2].co.y
            mdf_vert_y_h = highy_index[3].co.y

            src_vert_z_h = highz_index[2].co.z
            mdf_vert_z_h = highz_index[3].co.z

            #get average weights
            precision_x = ( mdf_vert_x_h - (src_vert_x_h + average_minimum_x) ) / 255 / 8
            precision_y = ( mdf_vert_y_h - (src_vert_y_h + average_minimum_y) ) / 31 / 64
            precision_z = ( mdf_vert_z_h - (src_vert_z_h + average_minimum_z) ) / 1023

            if precision_x == 0:
                precision_x = 0.0000001
            if precision_y == 0:
                precision_y = 0.0000001
            if precision_z == 0:
                precision_z = 0.0000001


            print( "\nMINS: " , average_minimum_x , average_minimum_y , average_minimum_z)

            print( "\ninter: " , precision_x )
            print( "inter: " , precision_y)
            print( "inter: " , precision_z)


            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( float_to_hex(precision_x) ) #x
            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( float_to_hex(precision_y) ) #y
            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( float_to_hex(precision_z) ) #z

            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( float_to_hex(average_minimum_x) ) #x
            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( float_to_hex(average_minimum_y) ) #y
            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( float_to_hex(average_minimum_z) ) #z

            for index in range( len(shapeVertArr) ):
                vert_arr = shapeVertArr[index]
                modif_arr = shapeModifArr[ index ]
                
                if shape_ID != "Basis":
                    for b in range( len(default_verts) ):

                        #collect-x
                        vert_x  = vert_arr[b].co.x
                        mvert_x = modif_arr[b].co.x
                        delta_x = mvert_x - vert_x

                        #collect-y
                        vert_y  = vert_arr[b].co.y
                        mvert_y = modif_arr[b].co.y
                        delta_y = mvert_y - vert_y

                        #collect-z
                        vert_z  = vert_arr[b].co.z
                        mvert_z = modif_arr[b].co.z
                        delta_z = mvert_z - vert_z

                        if math.isnan(mvert_x):
                            delta_x = 0.00001
                        if math.isnan(mvert_y):
                            delta_y = 0.00001
                        if math.isnan(mvert_z):
                            delta_z = 0.00001

                        if delta_x != 0 or delta_y != 0 or delta_z != 0:

                            ##format
                            var_x = ( ( delta_x - average_minimum_x ) / precision_x ) / 8
                            var_x = mbfd_dxgi_bin.roundUp( var_x , 1 )

                            var_y = ( delta_y - average_minimum_y ) / precision_y
                            var_y = mbfd_dxgi_bin.roundUp( var_y , 1 )
                            q_y , r_y = divmod( var_y , 63 )
                            var_y = q_y

                            if (r_y > 63):
                                r_y = 63

                            var_z = ( delta_z - average_minimum_z ) / precision_z
                            q_z , r_z = divmod( var_z , 1024 )
                            var_z = r_z + ( r_y * 1024 )


                            #out
                            dataBuffer.seek( 0 , 2 );dataBuffer.write( int(var_z).to_bytes(2, 'little' , signed=False)  ) #z
                            dataBuffer.seek( 0 , 2 );dataBuffer.write( int(var_y).to_bytes(1, 'little' , signed=False)  ) #y
                            dataBuffer.seek( 0 , 2 );dataBuffer.write( int(var_x).to_bytes(1, 'little' , signed=False)  ) #x
                            
                            #n
                            dataBuffer.seek( 0 , 2 );dataBuffer.write( int(0).to_bytes(1, 'little' , signed=False)  ) #n
                            dataBuffer.seek( 0 , 2 );dataBuffer.write( int(0).to_bytes(1, 'little' , signed=False)  ) #n
                            dataBuffer.seek( 0 , 2 );dataBuffer.write( int(0).to_bytes(1, 'little' , signed=False)  ) #n

                            dataBuffer.seek( 0 , 2 );dataBuffer.write( int(0).to_bytes(1, 'little' , signed=False)  ) #f

            #out           
            dataBuffer.seek(0)
            out = dataBuffer.read()

            #morph-size
            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( int( len( out )).to_bytes(4, 'little') )
            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( out ) 


            #########################
            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( int(blankEntry).to_bytes(4, 'little') ) #Item Dict
        else:
            blankEntry = mainText.index( '' )
            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( int(0).to_bytes(4, 'little') ) #morph Count
            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( int(blankEntry).to_bytes(4, 'little') ) #Item Dict

        if ( len (mdl_colors) ) != 0:                                                   #VColor ID
            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( int(0).to_bytes(4, 'little') )

        in_memory_sub.seek( 0 , 2 );in_memory_sub.write(b'ENDM\x00\x00\x00\x00')
        
        
    in_memory = io.BytesIO(b'MBfD')
    in_memory.seek( 0 , 2 );in_memory.write( (len(in_memory_sub.getvalue())).to_bytes(4, 'little') )    
    
    return in_memory.getvalue() , in_memory_sub.getvalue()

def get_binaryLODs( sceneData , mainText , hasRig ):
    subLods = 2
    if ( hasRig == 'false'):
        subLods = 2

    in_memory_sub = io.BytesIO(b'')
    in_memory_sub.seek( 0 , 2 );in_memory_sub.write( int(subLods).to_bytes(4, 'little') )
    modelFaces = sceneData[2]
    
    for z in range( subLods ):
        for a in range( len(sceneData[0]) ):

            meshVerts = len( sceneData[ 1 ][a] )
            faceCount = len( modelFaces[a][1] )
            
            if ( a == 0 ):
                in_memory_sub.seek( 0 , 2 );in_memory_sub.write( len(sceneData[0]).to_bytes(4, 'little') )
     
            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( int(a).to_bytes(2, 'little') )
            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( int(faceCount*3).to_bytes(4, 'little') )

            if ( meshVerts > 65535 ):
                for b in range( len( modelFaces[a][1] ) ):
                    in_memory_sub.seek( 0 , 2 );in_memory_sub.write( (modelFaces[a][1][b][1]-1).to_bytes(4, 'little') )
                    in_memory_sub.seek( 0 , 2 );in_memory_sub.write( (modelFaces[a][1][b][0]-1). to_bytes(4, 'little') )
                    in_memory_sub.seek( 0 , 2 );in_memory_sub.write( (modelFaces[a][1][b][2]-1).to_bytes(4, 'little') )
            else:
                for b in range( len( modelFaces[a][1] ) ):
                    in_memory_sub.seek( 0 , 2 );in_memory_sub.write( (modelFaces[a][1][b][1]-1).to_bytes(2, 'little') )
                    in_memory_sub.seek( 0 , 2 );in_memory_sub.write( (modelFaces[a][1][b][0]-1). to_bytes(2, 'little') )
                    in_memory_sub.seek( 0 , 2 );in_memory_sub.write( (modelFaces[a][1][b][2]-1).to_bytes(2, 'little') )   

            ##align    
            while ( len(in_memory_sub.getvalue() ) % 4 ):
                in_memory_sub.seek( 0 , 2 );in_memory_sub.write( int(0).to_bytes(1, 'little') )    
            
            #foot
            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( int(1).to_bytes(4, 'little') )
            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( int(a).to_bytes(4, 'little') ) 
            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( int(0).to_bytes(4, 'little') ) 
            in_memory_sub.seek( 0 , 2 );in_memory_sub.write( int(faceCount*3).to_bytes(4, 'little') )
            in_memory_sub.seek( 0 , 2 );in_memory_sub.write(b'ENDM\x00\x00\x00\x00')
        
    in_memory = io.BytesIO(b'LODs')
    in_memory.seek( 0 , 2 );in_memory.write( (len(in_memory_sub.getvalue())).to_bytes(4, 'little') )    
    
    return in_memory.getvalue() , in_memory_sub.getvalue()



def formatMDL( sceneData , outputFile , EXPORT_SHAPE_KEYS  ):
    
    mainText = formatTEXT( sceneData )
    
    #check if has rig
    hasRig = 'false'
    if ( len( sceneData[5] ) == 0 ): hasRig = 'false'
    else: hasRig = 'true'
    
    mdlHead = get_binaryMDLStart()
    textHead , textSub = get_binaryTEXT( mainText )
    mtlHead , mtlSub = get_binaryMTL( sceneData , mainText )
    mbfdHead , mbfdSub = get_binaryMBFD( sceneData , mainText , hasRig , EXPORT_SHAPE_KEYS )
    LODsHead , LODsSub = get_binaryLODs( sceneData , mainText , hasRig )
    
    if ( hasRig == 'true' ):
        boneHead , boneSub = get_binaryBONE( mainText , sceneData[5] , sceneData[6] , sceneData[7] , sceneData[9] )
    
    with open( outputFile , 'wb' ) as f:    
        f.write( mdlHead )
        f.write( textHead )
        f.write( textSub )
        if ( hasRig == 'true' ):
            f.write( boneHead )
            f.write( boneSub )
        f.write( mtlHead )
        f.write( mtlSub )
        f.write( mbfdHead )
        f.write( mbfdSub )
        f.write( LODsHead )
        f.write( LODsSub )
        f.write( b'END!\x00\x00\x00\x00' )

def clearSkeletonArray():
    global skelBones
    global skelBones_translate
    global skelBones_rotate
    global skelBones_parent

    skelBones = []
    skelBones_translate = []
    skelBones_rotate = []
    skelBones_parent = []

    return 

       
def mainExportMDL( filepath , args):

    EXPORT_EDGESPLIT = args[ "use_edgesplit" ]
    EXPORT_SHAPE_KEYS = args[ "export_blendshapes" ]

    clearSkeletonArray()
    init_config()

    sceneModels = getActiveModels()
    sceneData = []

    if ( len( sceneModels ) == 0 ):
        print( "No Objs sel" )
    else:
        sceneData = parseModels( sceneModels  , EXPORT_EDGESPLIT)
        print( "out" )

    for mesh in sceneData[0]:
        print( mesh )
    
    outFile = filepath

    print( "hello module")
     
    formatMDL( sceneData , outFile , EXPORT_SHAPE_KEYS )

    return
    
    
 