# vsCode module
from io_scene_mdl2022.mdl_utils_2022 import mdl_utils , scene_mdl_blender
from io_scene_mdl2022.mdl_utils_2022 import hpl_utils
import bpy
import configparser, os
import sys
from math import radians
import mathutils
import bpy , bmesh
from bpy import context
from mathutils import Vector, Quaternion
from mathutils import Euler, Matrix
from itertools import chain
import numpy as np
import math


global config_path
global user_path
global conf_path
global hasConfig

configBone = configparser.ConfigParser(); config_AtPt = configparser.ConfigParser()
user_path = bpy.utils.resource_path('USER'); conf_path = os.path.join(user_path, "config")
config_path = conf_path + "\mdl_skel.cache.ini" 
config_atpt_path = conf_path + "\mdl_AtPt.cache.ini" 
hasConfig = 'false'
hplList = None

def createArmature( armature, rig, Barr, rarr, tarr, Parr):

    modelBones = [] #store all bones for index values and matrix access
    modelTFMs = [] #to preserve matrix order, we store tfms as a list to preserve our formatted matrix

    for x in range( len( Barr ) ):
        name = Barr[x]; parentIndex = Parr[x]
        bone = armature.edit_bones.new( name )

        #create transform matrix based on given euler and translation
        rot = Euler( ( rarr[x][0] , rarr[x][1] , rarr[x][2] ), 'XYZ' )
        rotMat = formatMatrixArray( rot.to_matrix() , tarr[x] )

        #Evaluate matrix relative to parent if exists        
        if ( parentIndex != 65535 ):
            bone.parent = modelBones[ parentIndex ]
            parentMat = modelTFMs[ parentIndex ]
            rotMat = dotMatrixArrays( rotMat , parentMat )
            
        modelTFMs.append( rotMat )
        modelBones.append( bone )   

        #Set Bone's tail to access it's mat; Specify the final mat
        bone.tail = ( Vector((0.0, 1.0, 0.0)) * max(0.01, 1.0) )
        bone.matrix = reformatMatrix( rotMat )

        #Bone check to prevent culling
        if( bone.length < .00000001 ):                    
            bone.length = bone.length + .0001              
                
    ##rotate to match Blender's Z up
    rig.rotation_euler = (radians(-90), 0, 0 )



def createArmatureOld( armature, rig, Barr, rarr, tarr, Parr):

    modelBones = [] #store all bones for index values and matrix access
    modelTFMs = [] #to preserve matrix order, we store tfms as a list to preserve our formatted matrix

    for x in range( len( Barr ) ):
        name = Barr[x]; parentIndex = Parr[x]
        bone = armature.edit_bones.new( name )

        #create transform matrix based on given euler and translation
        rot = Euler( ( rarr[x][0] , rarr[x][1] , rarr[x][2] ), 'XYZ' )
        rotMat = formatMatrixArray( rot.to_matrix() , tarr[x] )

        #Evaluate matrix relative to parent if exists        
        if ( parentIndex != 65535 ):
            bone.parent = modelBones[ parentIndex ]
            parentMat = modelTFMs[ parentIndex ]
            rotMat = dotMatrixArrays( rotMat , parentMat )
            bone.tail = bone.parent.head

        modelTFMs.append( rotMat )
        modelBones.append( bone )   

        #Specify Bone's head according to our matrix
        bone.head = ( rotMat[3][0] , rotMat[3][1] ,
                                    rotMat[3][2] ) 

        #Bone check to prevent culling
        if( bone.length < .00000001 ):                    
            bone.length = bone.length + .0001              
                
    ##rotate to match Blender's Z up
    rig.rotation_euler = (radians(-90), 0, 0 )



def write_file( iniPath , configBf ):
    configBf.write( open( iniPath, 'w') )

def init_config_new( iniPath , configBf ):
    global user_path
    global conf_path
    global hasConfig

    user_path = bpy.utils.resource_path('USER')
    conf_path = os.path.join(user_path, "config")

    #clear old-cache
    if not os.path.exists(iniPath):
        write_file(iniPath,configBf)
        
    else:
        # Read File
        configBf['general'] = {}

        # remove temp files
        with open(iniPath, "w") as f:
            configBf.write(f)

#debug-notused
def getMDLFile():
    inputMDL = ""

    return inputMDL


def getMaxBoneVectors( blInds, blInfluence, vertices ):
    boneTotal = 0

    #Iterate through model vertices
    for vertex in range( vertices ):
        for i in range( blInfluence ):
            boneIndex = blInds[i+blInfluence*vertex]
            if boneIndex > boneTotal: boneTotal = boneIndex

    return boneTotal+1

def setupBoneStrings( textStrings, boneIDCeiling ):
    bnArrs = []
    for item in range( boneIDCeiling ):
        bnArrs.append( textStrings[item] )

    return bnArrs


def getVertexGroups( blInds , blWeights , textStrings , blInfluence ):
    blInds = list(chain(*blInds))
    blWeights = list(chain(*blWeights))

    vertices = int(len(blInds)/blInfluence)
    boneIDCeiling = getMaxBoneVectors( blInds, blInfluence, vertices )

    #initialize lists
    bnArrs = setupBoneStrings( textStrings, boneIDCeiling )
    vertArrs = [[] for i in range(boneIDCeiling)]
    wtArrs = [[] for i in range(boneIDCeiling)]

    #Iterate through model vertices
    for vertex in range( vertices ):

        #Iterate through every weight pertaining to 'vertex'
        for i in range( blInfluence ):
            boneIndex = blInds[i+blInfluence*vertex]
            vertWeight = blWeights[i+blInfluence*vertex]
                
            #append vertex to list
            vertArrs[ boneIndex ].append( vertex )
            wtArrs[ boneIndex ].append( vertWeight )

    return bnArrs , vertArrs , wtArrs

def formatMatrixArray( inputMatrix , posArr ):
    inputMatrix.invert()
    test = inputMatrix.to_4x4()
    test[3][0] = posArr[0]
    test[3][1] = posArr[1]
    test[3][2] = posArr[2]
    
    return test

def numpyToMatrix( array ):
    
    euler = Euler( ( 0 , 0 , 0 ), 'XYZ' )
    newMatrix = euler.to_matrix()
    newMatrix = newMatrix.to_4x4()
    
    newMatrix[0][0] = array[0][0]
    newMatrix[0][1] = array[0][1]
    newMatrix[0][2] = array[0][2]
    newMatrix[0][3] = array[0][3]
    
    newMatrix[1][0] = array[1][0]
    newMatrix[1][1] = array[1][1]
    newMatrix[1][2] = array[1][2]
    newMatrix[1][3] = array[1][3]
    
    newMatrix[2][0] = array[2][0]
    newMatrix[2][1] = array[2][1]
    newMatrix[2][2] = array[2][2]
    newMatrix[2][3] = array[2][3]
    
    newMatrix[3][0] = array[3][0]
    newMatrix[3][1] = array[3][1]
    newMatrix[3][2] = array[3][2]
    newMatrix[3][3] = array[3][3]

    return newMatrix

def dotMatrixArrays( inputMatrix , secMatrix ):
    a1 = [ inputMatrix[0][0] , inputMatrix[0][1] , inputMatrix[0][2] , inputMatrix[0][3] ]
    a2 = [ inputMatrix[1][0] , inputMatrix[1][1] , inputMatrix[1][2] , inputMatrix[1][3] ]
    a3 = [ inputMatrix[2][0] , inputMatrix[2][1] , inputMatrix[2][2] , inputMatrix[2][3] ]
    a4 = [ inputMatrix[3][0] , inputMatrix[3][1] , inputMatrix[3][2] , inputMatrix[3][3] ]
    
    b1 = [ secMatrix[0][0] , secMatrix[0][1] , secMatrix[0][2] , secMatrix[0][3] ]
    b2 = [ secMatrix[1][0] , secMatrix[1][1] , secMatrix[1][2] , secMatrix[1][3] ]
    b3 = [ secMatrix[2][0] , secMatrix[2][1] , secMatrix[2][2] , secMatrix[2][3] ]
    b4 = [ secMatrix[3][0] , secMatrix[3][1] , secMatrix[3][2] , secMatrix[3][3] ]
    
    a = np.array([ a1, a2, a3, a4 ])
    b = np.array([ b1, b2, b3, b4 ])
    out = a.dot(b)
    
    out = numpyToMatrix( out )
    
    return out

##  VColor Conversion -- Code by Grix    
def to_srgb(c):
    if c < 0.0031308:
        srgb = 0.0 if c < 0.0 else c * 12.92
    else:
        srgb = 1.055 * pow(c, 1.0 / 2.4) - 0.055

    return srgb;

#setup attachment point config
def setupAttachPointConfig( AtPtArray , sceneCollection, mainTEXT , rig ):

    for point in range( len(AtPtArray) ):
        attachIndex = AtPtArray[point][0]
        attachRoot = AtPtArray[point][1]
        attachJoint = AtPtArray[point][2]
        atpt_X = AtPtArray[point][3]
        atpt_Y = AtPtArray[point][4]
        atpt_Z = AtPtArray[point][5]

        AtPtSphereID = str(attachIndex)+"_"+str(attachRoot)+"_"+str(attachJoint)+"_Point" + str(point)
        AtPtSphereVector = str(atpt_X) + "," + str(atpt_Y) + "," + str(atpt_Z) 
        config_AtPt["general"][AtPtSphereID] = AtPtSphereVector

    # Overwrite cache
    with open(config_atpt_path, "w") as f:
        config_AtPt.write(f)

def reformatMatrix( inputMatrix  ):

    a1 = [ inputMatrix[0][0] , inputMatrix[1][0] , inputMatrix[2][0] , inputMatrix[3][0] ]
    a2 = [ inputMatrix[0][1] , inputMatrix[1][1] , inputMatrix[2][1] , inputMatrix[3][1] ]
    a3 = [ inputMatrix[0][2] , inputMatrix[1][2] , inputMatrix[2][2] , inputMatrix[3][2] ]
    a4 = [ inputMatrix[0][3] , inputMatrix[1][3] , inputMatrix[2][3] , inputMatrix[3][3] ]

    a = np.array([ a1, a2, a3, a4 ])
    a = numpyToMatrix(a)
    return a

def getRemappedVertexBuffer( remapVec , polygons , vtxMap ):
    #Utilize remap for correct vertice affect
    remappedVBf = []
    for face in polygons:
        faceIndices = []
        for vert in face.vertices:
            #query map for visibility flag
            vFlag = vtxMap[ remapVec[vert] ]
            faceIndices.append( vFlag )
        if (1 in faceIndices):
            remappedVBf.append( face.index )
    return remappedVBf

def createVisibilityMap(hplIndices,obj):
    verticeMap = []

    print("SETTING: " , hplIndices)
    #create visibility array
    for vert in obj.data.vertices:
        visibilityFlag = 0
        if (vert.index in hplIndices): visibilityFlag = 1
        verticeMap.append(visibilityFlag)

    return verticeMap

def createFaceMap( hplIndices, faceMapID, obj, remapVec ):
    # Get the mesh data from the object
    mesh = obj.data
    faces = mesh.polygons
    
    #We use specified indices otherwise we get incorrect results
    vtxMap = createVisibilityMap(hplIndices,obj)
    affectedPolys = getRemappedVertexBuffer(remapVec,faces,vtxMap)

    # Get the face map layer
    face_map_layer = obj.face_maps.new(name=faceMapID)
    face_map_layer.add(affectedPolys)
    
    return

def getSceneHPL( filepath ):
    global hplList
    hplPath = hpl_utils.findDirHPL(filepath)
    if hplPath:
        hplList = hpl_utils.readHPLFile(hplPath)
    else:
        return False

def appendHPLMaskToMesh( object, name, remapVec):

    #We will use only the highest lodLevel hpl mask
    highPolyLODs = hplList[0]
    
    for lod in highPolyLODs.objects:
        if (name == lod.name):
            #Collect Flag
            appearanceFlag = lod.sceneFlag

            #Format/Set Map ID
            if (appearanceFlag == 0): appearanceFlag = "_Ring"
            if (appearanceFlag == 1): appearanceFlag = "_Entrance"
            if (type(appearanceFlag) is int): 
                appearanceFlag = "_" + str(appearanceFlag)
            faceMapName = "HPL"+ appearanceFlag

            #Create Map
            print( "Loading HPL... " , faceMapName )
            print( "Source HPL... " , remapVec )
            try:
                createFaceMap( lod.visibility_vector, faceMapName, object, remapVec )
            except:
                print("Could not load .hpl")

    return
def importMDL( filepath , args ):

    IMPORT_LEGACY_RIG = args[ "use_false_bone_orientation"]
    IMPORT_SHAPE_KEYS = args[ "import_shape_keys" ]
    IMPORT_HPL = args["import_hpl"]
    
    # initialize HPL if enabled (or if exists)
    if (IMPORT_HPL): 
        if ( not getSceneHPL(filepath) ): IMPORT_HPL = False

    # get new config
    init_config_new( config_path , configBone )
    init_config_new( config_atpt_path, config_AtPt )

    # get file input
    mainMDL = filepath
    print( "\nMain MDL File is: " + mainMDL )
    
    # read MDL file
    sceneData = mdl_utils.readMain( mainMDL , IMPORT_SHAPE_KEYS )
    modelCount = len( sceneData[2] )

    ## check error; do something
    if ( modelCount == 0 ):
        print( "No Models Loaded" )
    else:
        print( "\n" )


    
   # Make New Collection
        collName = os.path.basename(mainMDL).split(".")[0]
        mdlCollection = bpy.data.collections.new( collName )
        bpy.context.scene.collection.children.link(mdlCollection)

        #Write Rig
        isSkinned = 'true'
        if ( len (sceneData[3]) == 0 ):
            isSkinned = False
        else:   
            # Append bone structure
            tarr = sceneData[3][0]; rarr = sceneData[3][1]
            Barr = sceneData[3][2]; Parr = sceneData[3][3]
            item = mdlCollection

            armature = bpy.data.armatures.new("Skeleton")
            rig = bpy.data.objects.new("Skeleton", armature)
            mdlCollection.objects.link(rig)
            armature.display_type = "WIRE"
            
            bpy.context.view_layer.objects.active = rig
            bpy.ops.object.editmode_toggle()
            
            #create model's skeleton
            if (IMPORT_LEGACY_RIG ): createArmatureOld( armature, rig, Barr, rarr, tarr, Parr)
            else: createArmature( armature, rig, Barr, rarr, tarr, Parr)
                  
        #Write AtPt
        attachmentPoints = sceneData[5]
        if( len(attachmentPoints) != 0): setupAttachPointConfig( attachmentPoints, mdlCollection, sceneData[4], armature )


    # append mesh buffer data to scene
    for x in range( len(sceneData[2][1]) ):
        LODs = sceneData[2][0][x]
        meshID = sceneData[2][1][x]
        mesh_name = sceneData[1][2][x]
        modelShapeData = sceneData[1][3][x]
        hasColor = 'false'

        mainText = sceneData[4]
        print ("MESH ID: " + meshID)
        modelIndex = x
        subSkinned = 'true'

        try:
            blInds = sceneData[1][0][x].index('BLENDINDICES')
        except:
            subSkinned = 'false'
            isSkinned = 'false'
            pass

        verts = sceneData[1][0][x].index('POSITION')
        verts = sceneData[1][1][x][verts]
        
        norms = sceneData[1][0][x].index('NORMAL')
        norms = sceneData[1][1][x][norms]

        if 'COLOR' in sceneData[1][0][x]:
            hasColor = 'true'
            vcolors = sceneData[1][0][x].index('COLOR')
            vcolors = sceneData[1][1][x][vcolors]
            print("mesh has color")
        
        if isSkinned == 'true':
            if ( subSkinned != 'false'):
                blInds = sceneData[1][0][x].index('BLENDINDICES')
                blInds = sceneData[1][1][x][blInds]
                
                blWeights = sceneData[1][0][x].index('BLENDWEIGHTS')
                blWeights = sceneData[1][1][x][blWeights]
            
        meshUVCount = sceneData[1][0][x].count('TEXCOORD')
        UVindices = [i for i, j in enumerate(sceneData[1][0][x]) if j == "TEXCOORD"]
        
#        bn vt wt
        if isSkinned == 'true':
            if ( subSkinned != 'false'):
                rigData = getVertexGroups( blInds , blWeights , mainText, 4 )

        for j in range( len(norms) ):
            if ( len(norms[j]) > 3 ):
                norms[j].pop()
                

        new_mesh = bpy.data.meshes.new( mesh_name )
        new_mesh.from_pydata(verts, [], LODs)
        new_mesh.use_auto_smooth = True
        new_mesh.create_normals_split()

        
        loopSet = list( range( len( new_mesh.loops ) ) ) 

        for loop in new_mesh.loops:
            loop_vIdx = loop.vertex_index
            loop.normal = norms[ loop_vIdx ]
            loopSet[ loop.index ] = loop.normal
        new_mesh.normals_split_custom_set( loopSet )
        new_mesh.update()

        #Use Appearance Flags
        modelFlags = sceneData[1][0][x].index("Mesh Visibility Flags")
        modelFlags = sceneData[1][1][x][modelFlags]
        if( modelFlags[0] == 129 ): meshID += "!"
        if( modelFlags[0] == 130 ): meshID += "$"
        if( modelFlags[0] == 385 ): meshID += "!"
        
        #mesh_object
        new_object = bpy.data.objects.new( meshID , new_mesh )
        for f in new_object.data.polygons:
            f.use_smooth = True
        new_object.data.normals_split_custom_set( loopSet )

        #Append HPL Mask
        if(IMPORT_HPL and hplList):
            remapVector = sceneData[1][0][x].index("Vertex Remap")
            remapVector = sceneData[1][1][x][remapVector]
            appendHPLMaskToMesh(new_object,mesh_name, remapVector)


        if hasColor == 'true':
            new_mesh.vertex_colors.new()
            color_layer = new_mesh.vertex_colors.active
            i = 0
            for poly in new_mesh.polygons:
                for v_ix, l_ix in zip(poly.vertices, poly.loop_indices):
                    vert_color = vcolors[v_ix]
                    color_layer.data[i].color = ( to_srgb(vert_color[0]), to_srgb(vert_color[1]), to_srgb(vert_color[2]), vert_color[3])
                    i += 1
                
        if ( meshUVCount != 0 ):
            for h in range( meshUVCount ):
                meshUV = UVindices[h]
                meshUV = sceneData[1][1][modelIndex][meshUV]
                
                UVName = (meshID + "_UV" + str(h))
                
                #CreateNewUVLayer
                new_uv = new_object.data.uv_layers.new(name=UVName)
                           
                for face in new_mesh.polygons:
                    for vert_idx, loop_idx in zip(face.vertices, face.loop_indices):
                        new_uv.data[loop_idx].uv = [ meshUV[vert_idx][0] , -1.0 * (meshUV[vert_idx][1] - 1.0) ]

        if isSkinned == 'true':    
            #add rigData
            for grp in range( len(rigData[0]) ):
                
                boneName = rigData[0][grp]
                vertex_group_data = rigData[1][grp]
                weight_group_data = rigData[2][grp]
                
                vertGroup = new_object.vertex_groups.new(name=boneName)

                for x in range( len( vertex_group_data ) ):
                    var_0 = vertex_group_data[x]
                    weight = weight_group_data[x]
                    vertGroup.add(  [var_0] , weight , 'REPLACE')
                    

        if isSkinned == 'true':         
            # add object to scene collection
            new_object.parent = rig
            modifier = new_object.modifiers.new(type='ARMATURE', name="Armature")
            modifier.object = rig
        else:
            new_object.rotation_euler[0] = radians(-90)
        
        mdlCollection.objects.link(new_object)

        ##shape-config
        if (IMPORT_SHAPE_KEYS == False):
            mesh_name = "notUsed"  #why? 

        if( mesh_name != "notUsed"):

            morphShapes = len( modelShapeData )

            if morphShapes != 0:
                
                #add base shape key
                verts = new_object.data.vertices
                sk_basis = new_object.shape_key_add( name="Basis", from_mix=False )
                sk_basis.interpolation = 'KEY_LINEAR'
                new_object.data.shape_keys.use_relative = True

                for item in range( morphShapes ):
                    trans_list = modelShapeData[item][0]
                    morph_deltas = modelShapeData[item][1]
                    morph_ID = modelShapeData[item][2]
                    
                    changedVerts = len( trans_list )
                    
                    #add new shape key
                    sk = new_object.shape_key_add( name=morph_ID, from_mix=False )
                    sk.interpolation = 'KEY_LINEAR'
                    
                    var = 0
                    try:
                        for vert in trans_list:
                            vert_delta = morph_deltas[var]
                            vert_x = new_object.data.vertices[vert].co.x
                            vert_y = new_object.data.vertices[vert].co.y
                            vert_z = new_object.data.vertices[vert].co.z

                            sk.data[vert].co.x = vert_x + vert_delta[0]
                            sk.data[vert].co.y = vert_y + vert_delta[1]
                            sk.data[vert].co.z = vert_z + vert_delta[2]
                            # print( vert_delta )
                            
                            var += 1
                    except:
                        pass

    try:
        #frame mesh in viewport
        bpy.ops.view3d.view_all()

        #toggle outline shading
        bpy.context.space_data.shading.show_object_outline = False
        
    except:
        print( "OPS ERROR, CONTEXT SWITCHING... ")
        print( bpy.context.area.type )
        
        context = bpy.context
        for screen in bpy.data.screens:
            for area in (a for a in screen.areas if a.type == 'VIEW_3D'):
                region = next((region for region in area.regions if region.type == 'WINDOW'), None)
                if region is not None:
                    override = {'area': area, 'region': region} 
                    bpy.ops.view3d.view_all(override)

                    #toggle outline shading
                    area.spaces.active.shading.show_object_outline = False
    
        pass

    return