from cmath import pi
from glob import glob
from http.client import ImproperConnectionState
import io
from pkgutil import get_loader
from pyexpat import model
from io_scene_mdl2022.mdl_utils_2022 import mbfd_dxgi_bin
import math
import numpy as np
from io_scene_mdl2022.eulerangles import euler2matrix
import struct


class MBfD_model(object):
    def __init__(self, property, format, type):
        self.property = type
        self.format = format
        self.type = type

global modelCount
global MTLsList
global MBfDList
global LODsList
global modelVerts
global BONE    
global AtPtList

modelCount = 0
MTLsList = []
MBfDList = []
LODsList = []
modelVerts = []
BONE = []

def getRemapVertices( modelBuffer, meshVerts ):
    remapVector = []
    for x in range( meshVerts ):
        index = int.from_bytes( (modelBuffer.read(4) ) , "little" )
        remapVector.append( index )
    return remapVector

def DELTA_ELEMENTS_FROM_BIT( bit_value , floor ):

    deltaLookupTable = { '0': [],    
    '1': [0], '2': [1], '3': [0,1], '4': [2],
    '5': [0,2], '6': [1,2], '7': [0,1,2], '8': [3],
    '9': [0,3], 'A': [1,3], 'B': [0,1,3], 'C': [2,3],
    'D': [0,2,3], 'E': [1,2,3], 'F': [0,1,2,3]
    }

    bit_value = deltaLookupTable[bit_value]
    bit_value = [elem + (floor * 4) for elem in bit_value]

    return bit_value

def MORPH_DWORD_TO_ARRAY( input_var , floor ):

    varBits = f"{input_var:08X}"

    #split every byte
    n = 2
    varBits = [varBits[i:i+n] for i in range(0, len(varBits), n)]

    #create vert list
    weightedVertices = []

    #collect affected verts
    index = 0
    for x in range( len(varBits) ):
        indexBit = varBits[x][::-1]

        for v in range( len(indexBit) ):
            quad = DELTA_ELEMENTS_FROM_BIT( indexBit[v] , index ) # 0x D
            index += 1

            [ weightedVertices.append(elem+floor) for elem in quad ]

    return weightedVertices

def readMain( mainMDL , IMPORT_SHAPE_KEYS ):
    
    global modelCount
    global MTLsList
    global MBfDList
    global LODsList
    global modelVerts
    global BONE   
    global AtPtList
    global readShapes
    readShapes = IMPORT_SHAPE_KEYS

    modelCount = 0
    MTLsList = []
    MBfDList = []
    LODsList = []
    modelVerts = []
    BONE = []
    AtPtList = []

    file = open( mainMDL ,"rb")
    mdlMagic = file.read(4)

    #verify file integrity
    if ( mdlMagic != b'MDL!' ):
        print( "File is not MDL.")
        file.close(); return
    else:
        print( "File is valid MDL\n\n")

    #skip internal hash
    file.seek( 32 )

    #get TEXT data
    textMagic = file.read(4)
    textSize = int.from_bytes( ( file.read(4) ) , "little" )
    textArr = io.BytesIO( file.read( textSize ) )
    mainTEXT = get_TEXT( textArr , textSize )

    # interpret data blocks
    parseModelBlocks( file , mainTEXT )
    parseModelBlocks( file , mainTEXT )
    parseModelBlocks( file , mainTEXT )
    parseModelBlocks( file , mainTEXT )
    parseModelBlocks( file , mainTEXT )

    ## MBFD at 0 is data Types
    ## MBFD at 1 is data Sets
    ## LODs at 0 is face Sets
    ## LODs at 1 is mesh IDs

    file.close()

    return MTLsList , MBfDList , LODsList , BONE , mainTEXT ,  AtPtList

def parseModelBlocks( file , textStrings ):
    global modelCount
    global BONE
    global AtPtList
    global MTLsList
    global MBfDList
    global LODsList


    #get general data
    blockMagic = file.read(4)
    blockSize = int.from_bytes( ( file.read(4) ) , "little" )
    blockArr = io.BytesIO( file.read( blockSize ) )

    if (blockMagic == b'BONE'):
        print("BONE")
        BONE = get_BONE( blockArr , blockSize , textStrings )
    if (blockMagic == b'AtPt'):
        print("AtPt")
        AtPtList = get_AtPt( blockArr , blockSize , textStrings )
    if (blockMagic == b'MTL!'):
        print("MTL!")
        MTLsList = get_MTL( blockArr , blockSize )
    if (blockMagic == b'MBfD'):
        print("MBfD")
        MBfDList = get_MBfD( blockArr , blockSize , textStrings )
    if (blockMagic == b'LODs'):
        print("LODs")
        LODsList = get_LODs( blockArr , blockSize , textStrings )   
    

def get_AtPt(  textBuffer , textSize , textStrings ):
    attachmentPoints = int.from_bytes( (textBuffer.read(4) ) , "little" )
    attachmentArr = []

    for pointIdx in range( attachmentPoints ):
        attachIndex = int.from_bytes( (textBuffer.read(2) ) , "little" )
        attachRoot = int.from_bytes( (textBuffer.read(2) ) , "little" )
        attachJoint = int.from_bytes( (textBuffer.read(4) ) , "little" )

        atpt_X = int.from_bytes( (textBuffer.read(4) ) , "little" ); atpt_X = struct.unpack('f', struct.pack('I', atpt_X))[0]
        atpt_Y = int.from_bytes( (textBuffer.read(4) ) , "little" ); atpt_Y = struct.unpack('f', struct.pack('I', atpt_Y))[0]
        atpt_Z = int.from_bytes( (textBuffer.read(4) ) , "little" ); atpt_Z = struct.unpack('f', struct.pack('I', atpt_Z))[0]
        nullDWORD = int.from_bytes( (textBuffer.read(4) ) , "little" )

        attachmentArr.append( [ attachIndex, attachRoot, attachJoint, atpt_X, atpt_Y, atpt_Z ] )

    return attachmentArr  

def get_TEXT( textBuffer , textSize ):
    stringCount = int.from_bytes( (textBuffer.read(4) ) , "little" )
    stringOffList = []
    textStrings = []

    # collect char offset for all strings
    for x in range(stringCount):
        stringOffset = int.from_bytes( (textBuffer.read(4) ) , "little" )
        stringOffList.append( stringOffset )

    # collect actual strings
    for x in range(stringCount):
        textBuffer.seek( 4 + (4*stringCount) + stringOffList[x] , 0 )
        stringArray = b''; bitValue = b'\xFF'

        while ( bitValue != b'\x00' ):
            bitValue = textBuffer.read(1)
            stringArray += bitValue

        textStrings.append( stringArray[:-1].decode( 'utf-8' ) )

    ## ---print( textStrings )--
    print("TEXT Count is:" , str( stringCount ) )

    # return TEXTlist
    return textStrings

def get_BONE(   boneBuffer , boneSize , textStrings ):
    boneCount = int.from_bytes( (boneBuffer.read(4) ) , "little" )

    tarr = []
    rarr = []
    Barr = []
    Parr = []

    for i in range(boneCount):
        boneInd = int.from_bytes( (boneBuffer.read(2) ) , "little" )
        bonePar = int.from_bytes( (boneBuffer.read(2) ) , "little" )
        boneName= textStrings[boneInd]
        parent = bonePar

        #get point3 values
        tx = int.from_bytes( (boneBuffer.read(4) ) , "little" )
        tx = mbfd_dxgi_bin.DWORDFloat( tx )
        ty = int.from_bytes( (boneBuffer.read(4) ) , "little" )
        ty = mbfd_dxgi_bin.DWORDFloat( ty )
        tz = int.from_bytes( (boneBuffer.read(4) ) , "little" )
        tz = mbfd_dxgi_bin.DWORDFloat( tz )

        rx = int.from_bytes( (boneBuffer.read(4) ) , "little" )
        rx = mbfd_dxgi_bin.DWORDFloat( rx )
        ry = int.from_bytes( (boneBuffer.read(4) ) , "little" )
        ry = mbfd_dxgi_bin.DWORDFloat( ry )
        rz = int.from_bytes( (boneBuffer.read(4) ) , "little" )
        rz = mbfd_dxgi_bin.DWORDFloat( rz )


        tarr.append( [tx,ty,tz] )
        rarr.append( [rx , ry , rz ] )
        Barr.append( boneName )
        Parr.append( parent )
    
    return tarr , rarr , Barr , Parr


def get_MTL(  mtlBuffer , mtlSize ):

    mtlCount = int.from_bytes( (mtlBuffer.read(4) ) , "little" )
    mtlIndices = []

    for x in range(mtlCount):
        mtlIndex = int.from_bytes( (mtlBuffer.read(4) ) , "little" )
        mtlIndices.append( mtlIndex )

    return mtlIndices
    

def get_MBfD( modelBuffer , MBfDSize , mainTEXT ):
    global modelCount
    modelCount = int.from_bytes( (modelBuffer.read(4) ) , "little" )
    modelDataSets = []
    modelDataTypes = []
    modelNames = []
    modelMorphs = []

    print ( "Model Count is: " + str(modelCount) )

    # iterate over every sub model
    for x in range( modelCount ):
        submodelDataSets = []
        submodelDataTypes = []

        if (x > 0):
            tempBuf = b''
            while (tempBuf != b'ENDM'):
                tempBuf = (modelBuffer.read(4))
            modelBuffer.seek( modelBuffer.tell() + 4 )
                   
        # read basic mbfd data
        modelName = mainTEXT[ int.from_bytes( (modelBuffer.read(4) ) , "little" ) ]
        modelAppearanceFlag00 =  int.from_bytes( (modelBuffer.read(4) ) , "little" )
        modelAppearanceFlag01 =  int.from_bytes( (modelBuffer.read(2) ) , "little" )

        # skip model hash
        modelBuffer.seek( modelBuffer.tell() + 28 )

        # read general model data
        modelVertCount = int.from_bytes( (modelBuffer.read(4) ) , "little" )
        modelDataBlocks = int.from_bytes( (modelBuffer.read(4) ) , "little" )
        modelVerts.append( modelVertCount )

        ## collect subData blocks
        for i in range( modelDataBlocks ):

            subBlockData = []

            # skip block hash --these are always const
            modelBuffer.seek( modelBuffer.tell() + 12 )
            blockProperty = mainTEXT[ int.from_bytes( (modelBuffer.read(2) ) , "little" ) ]
            blockFormat = mainTEXT[ int.from_bytes( (modelBuffer.read(2) ) , "little" ) ]
            blockType = mainTEXT[ int.from_bytes( (modelBuffer.read(4) ) , "little" ) ]

            # skip if init block
            if (i == 0):
                modelBuffer.seek( modelBuffer.tell() + 2 )

            # conditionals for data types
            if ( blockFormat == "R32_G32_B32" ):
                outBuffer = io.BytesIO( modelBuffer.read( 12 * modelVertCount ) )
                subBlockData = mbfd_dxgi_bin.getData_R32_G32_B32( outBuffer , modelVertCount , blockType )

            if ( blockFormat == "R8_G8_B8_A8" ):
                outBuffer = io.BytesIO( modelBuffer.read( 4 * modelVertCount ) )
                subBlockData = mbfd_dxgi_bin.getData_R8_G8_B8_A8( outBuffer , modelVertCount , blockType )

            if ( blockFormat == "R16_G16_B16_A16" ):
                outBuffer = io.BytesIO( modelBuffer.read( 8 * modelVertCount ) )
                subBlockData = mbfd_dxgi_bin.getData_R16_G16_B16_A16( outBuffer , modelVertCount , blockType )
                
            if ( blockFormat == "R32_G32" ):
                outBuffer = io.BytesIO( modelBuffer.read( 8 * modelVertCount ) )
                subBlockData = mbfd_dxgi_bin.getData_R32_G32( outBuffer , modelVertCount , blockType )

            if ( blockFormat == "R8" ):
                outBuffer = io.BytesIO( modelBuffer.read( modelVertCount ) )
                subBlockData = mbfd_dxgi_bin.getData_R8( outBuffer , modelVertCount , blockType )
                alignPos = mbfd_dxgi_bin.checkAlignment( modelBuffer.tell() )
                modelBuffer.seek( alignPos )

            submodelDataTypes.append( blockProperty )
            submodelDataSets.append( subBlockData )
            
            # print( blockProperty + blockFormat + blockType )

        ###  POST-data blocks
        
        #------------------------------------------------------
        #------------------------------------------------------
        #------------------------------------------------------

        ##  Vertex Remapping
        remapVector = getRemapVertices( modelBuffer , modelVertCount )
        submodelDataTypes.append( "Vertex Remap" )
        submodelDataSets.append( remapVector )
        

        #Flags
        modelFlags = [ modelAppearanceFlag00 , 00 ]
        submodelDataTypes.append( "Mesh Visibility Flags" )
        submodelDataSets.append( modelFlags )

        ##  MorphCount
        modelMorphCount = int.from_bytes( (modelBuffer.read(4) ) , "little" )
        morphIDs = []
        morph_transforms = []

        #Toggle Import
        if (readShapes == False):
            modelMorphCount == 0

        if ( modelMorphCount != 0 and readShapes != False ):

            ## collect morph IDs
            for x in range( modelMorphCount ):
                morphIndex = int.from_bytes( (modelBuffer.read(2) ) , "little" )
                morphIDs.append( mainTEXT[ morphIndex ] )

            #align if morph counts misalign buffer
            alignPos = mbfd_dxgi_bin.checkAlignment( modelBuffer.tell() )
            modelBuffer.seek( alignPos )
            
            ## morph verts
            morphVerts = int.from_bytes( (modelBuffer.read(4) ) , "little" )

            ## delta floats 0 (transform)
            morph_float_x = mbfd_dxgi_bin.DWORDFloat( int.from_bytes( (modelBuffer.read(4) ) , "little" ) )
            morph_float_y = mbfd_dxgi_bin.DWORDFloat( int.from_bytes( (modelBuffer.read(4) ) , "little" ) )
            morph_float_z = mbfd_dxgi_bin.DWORDFloat( int.from_bytes( (modelBuffer.read(4) ) , "little" ) )

            ## delta floats 1 (pre-transform)
            morph_float_x1 = mbfd_dxgi_bin.DWORDFloat( int.from_bytes( (modelBuffer.read(4) ) , "little" ))
            morph_float_y1 = mbfd_dxgi_bin.DWORDFloat( int.from_bytes( (modelBuffer.read(4) ) , "little" ))
            morph_float_z1 = mbfd_dxgi_bin.DWORDFloat( int.from_bytes( (modelBuffer.read(4) ) , "little" ))

            ## morph section size
            morphSize = int.from_bytes( (modelBuffer.read(4) ) , "little" )
            filePos = modelBuffer.tell()
            morph_transforms = list( range(modelMorphCount) )

            if (modelName != "notUsed"):
                for item in range( modelMorphCount ): 
                    x = 0
                    trans_list = []
                    morph_deltas = []
                    morph_ID = morphIDs[item]

                    while x < morphVerts:
                        morphedVerts = int.from_bytes( (modelBuffer.read(4) ) , "big" )
                        morphVOffset = int.from_bytes( (modelBuffer.read(4) ) , "little" )

                        morphedVerts = MORPH_DWORD_TO_ARRAY( morphedVerts , x )
                        [trans_list.append( elem ) for elem in morphedVerts]  

                        #cache old pos
                        prePos = modelBuffer.tell()
                        modelBuffer.seek( filePos + (morphVOffset*4) )
                        vert_deltas = []

                        #collect deltas
                        for vIdx in range( len(morphedVerts) ):
                            #z
                            #y
                            #x

                            delta_vert_z = int.from_bytes( (modelBuffer.read(2) ) , "little", signed=False )

                            delta_vert_y = int.from_bytes( (modelBuffer.read(1) ) , "little", signed=False )
                            
                            delta_vert_x = int.from_bytes( (modelBuffer.read(1) ) , "little", signed=False )
                            
                            delta_vert_unk1 = int.from_bytes( (modelBuffer.read(2) ) , "little" )
                            
                            delta_vert_n = int.from_bytes( (modelBuffer.read(2) ) , "little" )

                            
                            #z nibble
                            q_z , r_z = divmod(delta_vert_z, 1024)
                            delta_vert_z = r_z

                            #y nibble
                            q, r = divmod(delta_vert_y, 32)
                            delta_vert_y = r

                            vert_deltas.append(
                                [( morph_float_x1 +  8 *  ( morph_float_x )  * delta_vert_x ), 
                                 ( morph_float_y1 + ( ( q_z + (64 * delta_vert_y) ) * morph_float_y ) ) , 
                                 ( morph_float_z1 + ( delta_vert_z * morph_float_z ) ) ,
                                delta_vert_n ]  )

                        #next
                        modelBuffer.seek( prePos )
                        [morph_deltas.append( elem ) for elem in vert_deltas]  
                        x+=32
                    
                    morph_transforms[ item ] = [ trans_list , morph_deltas , morph_ID ]

        modelDataTypes.append( submodelDataTypes )
        modelDataSets.append( submodelDataSets )
        modelNames.append( modelName )
        modelMorphs.append( morph_transforms )

        print ( "\nSubModel: " + str(modelName) + "\n Vert Count: " + str(modelVertCount) )


    return modelDataTypes , modelDataSets , modelNames , modelMorphs

    
def get_LODs( modelBuffer , MBfDSize , mainTEXT ):
    global modelCount
    mainLOD = []
    mainMeshID = []

    for i in range( modelCount ):
        subLODarr = []
        outOff = modelBuffer.tell()
        model_vertices = modelVerts[i]

        if (i == 0):
            subLODs = int.from_bytes( (modelBuffer.read(4) ) , "little" )
        LODCount = int.from_bytes( (modelBuffer.read(4) ) , "little" )
        LODindex = int.from_bytes( (modelBuffer.read(2) ) , "little" )
        faceCount = int.from_bytes( (modelBuffer.read(4) ) , "little" )

        if model_vertices > 65535:
            for x in range( int(faceCount/3) ):
                fx = int.from_bytes( (modelBuffer.read(4) ) , "little" )
                fy = int.from_bytes( (modelBuffer.read(4) ) , "little" )
                fz = int.from_bytes( (modelBuffer.read(4) ) , "little" )

                subLODarr.append( [fy , fx , fz] )
        else:
            for x in range( int(faceCount/3) ):
                fx = int.from_bytes( (modelBuffer.read(2) ) , "little" )
                fy = int.from_bytes( (modelBuffer.read(2) ) , "little" )
                fz = int.from_bytes( (modelBuffer.read(2) ) , "little" )

                subLODarr.append( [fy , fx , fz] )

        mainLOD.append( subLODarr )

        #Check Alignment
        alignedOff = mbfd_dxgi_bin.checkAlignment( modelBuffer.tell() )
        modelBuffer.seek( alignedOff )
        
        #Iterate through per face material entries
        meshID = ""
        materialEntries = int.from_bytes( (modelBuffer.read(4) ) , "little" )
        for m in range(materialEntries):
            mtlIdx = int.from_bytes( (modelBuffer.read(4) ) , "little" )
            matFaceBegin = int.from_bytes( (modelBuffer.read(4) ) , "little" )
            matFaceLength = int.from_bytes( (modelBuffer.read(4) ) , "little" )
            #We will only assign the last material
            meshID = mainTEXT[ MTLsList[ mtlIdx ] ]

        #seek END signature and loop
        tempBuf = b''
        while (tempBuf != b'ENDM'):
            tempBuf = (modelBuffer.read(4))
        
        mainMeshID.append( meshID )

    return mainLOD , mainMeshID
